from __future__ import annotations

from typing import List, Union, overload, Optional, TYPE_CHECKING

from pymeos_cffi import floatset_in, floatset_make, floatset_out, \
    floatset_start_value, floatset_end_value, floatset_value_n, \
    floatset_values, contains_floatset_float, intersection_floatset_float, \
    intersection_set_set, left_floatset_float, overleft_floatset_float, \
    right_floatset_float, overright_floatset_float, minus_floatset_float, \
    minus_set_set, union_set_set, union_floatset_float, floatset_shift_scale, \
    minus_float_floatset, distance_floatset_float

from .floatspan import FloatSpan
from .floatspanset import FloatSpanSet
from ..base import Set

if TYPE_CHECKING:
    from .intset import IntSet


class FloatSet(Set[float]):
    """
    Class for representing a set of text values.

    ``TextSet`` objects can be created with a single argument of type string as
    in MobilityDB.

        >>> FloatSet(string='{1, 3, 56}')

    Another possibility is to create a ``TextSet`` object from a list of
    strings or floats.

        >>> FloatSet(elements=[1, '2', 3, '56'])


    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'floatset'

    _parse_function = floatset_in
    _parse_value_function = float
    _make_function = floatset_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------

    def __str__(self, max_decimals: int = 15):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            floatset_out
        """
        return floatset_out(self._inner, max_decimals)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self) -> FloatSpanSet:
        """
        Returns a SpanSet that contains a Span for each element in ``self``.

        Returns:
            A new :class:`FloatSpanSet` instance

        MEOS Functions:
            set_to_spanset
        """

        return FloatSpanSet(_inner=super().to_spanset())

    def to_span(self) -> FloatSpan:
        """
        Returns a span that encompasses ``self``.

        Returns:
            A new :class:`FloatSpan` instance

        MEOS Functions:
            set_span
        """
        return FloatSpan(_inner=super().to_span())

    def to_intset(self) -> IntSet:
        """
        Converts ``self`` to an :class:`IntSet` instance.

        Returns:
            A new :class:`IntSet` instance
        """
        from .intset import IntSet
        return IntSet(elements=[int(x) for x in self.elements()])

    # ------------------------- Accessors -------------------------------------

    def start_element(self) -> float:
        """
        Returns the first element in ``self``.

        Returns:
            A :class:`float` instance

        MEOS Functions:
            floatset_start_value
        """
        return floatset_start_value(self._inner)

    def end_element(self) -> float:
        """
        Returns the last element in ``self``.

        Returns:
            A :class:`float` instance

        MEOS Functions:
            floatset_end_value
        """
        return floatset_end_value(self._inner)

    def element_n(self, n: int) -> float:
        """
        Returns the ``n``-th element in ``self``.

        Args:
            n: The 0-based index of the element to return.

        Returns:
            A :class:`float` instance

        MEOS Functions:
            floatset_value_n
        """
        super().element_n(n)
        return floatset_value_n(self._inner, n + 1)

    def elements(self) -> List[float]:
        """
        Returns the elements in ``self``.

        Returns:
            A list of :class:`float` instances

        MEOS Functions:
            floattset_values
        """
        elems = floatset_values(self._inner)
        return [elems[i] for i in range(self.num_elements())]

    # ------------------------- Transformations ------------------------------------

    def shift(self, delta: float) -> FloatSet:
        """
        Returns a new ``FloatSet`` instance with all elements shifted by ``delta``.

        Args:
            delta: The value to shift by.

        Returns:
            A new :class:`FloatSet` instance

        MEOS Functions:
            floatset_shift_scale
        """
        return self.shift_scale(delta, None)

    def scale(self, new_width: float) -> FloatSet:
        """
        Returns a new ``FloatSet`` instance with all elements scaled to so that the encompassing
        span has width ``new_width``.

        Args:
            new_width: The new width.

        Returns:
            A new :class:`FloatSet` instance

        MEOS Functions:
            floatset_shift_scale
        """
        return self.shift_scale(None, new_width)

    def shift_scale(self, delta: Optional[float], new_width: Optional[float]) -> FloatSet:
        """
        Returns a new ``FloatSet`` instance with all elements shifted by ``delta`` and scaled to so that the
         encompassing span has width ``new_width``.

        Args:
            delta: The value to shift by.
            new_width: The new width.

        Returns:
            A new :class:`FloatSet` instance

        MEOS Functions:
            floatset_shift_scale
        """
        return FloatSet(
            _inner=floatset_shift_scale(self._inner, delta, new_width, delta is not None, new_width is not None))

    # ------------------------- Topological Operations --------------------------------

    def contains(self, content: Union[FloatSet, float]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set, contains_floatset_float
        """
        if isinstance(content, float):
            return contains_floatset_float(self._inner, content)
        else:
            return super().contains(content)

    # ------------------------- Position Operations --------------------------------

    def is_left(self, content: Union[FloatSet, float]) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            left_set_set, left_floatset_float
        """
        if isinstance(content, float):
            return left_floatset_float(self._inner, content)
        else:
            return super().is_left(content)

    def is_over_or_left(self, content: Union[FloatSet, float]) -> bool:
        """
        Returns whether ``self`` is to the left of ``other`` allowing overlap.
        That is, ``self`` ends before ``other`` ends (or at the same value).

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            overleft_set_set, overleft_floatset_float
        """
        if isinstance(content, float):
            return overleft_floatset_float(self._inner, content)
        else:
            return super().is_over_or_left(content)

    def is_right(self, content: Union[FloatSet, float]) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``. That is,
        ``self`` starts before ``other`` ends.

        Args:
            content: object to compare with

        Returns:
            True if is rigth, False otherwise

        MEOS Functions:
            right_set_set, right_floatset_float
        """
        if isinstance(content, float):
            return right_floatset_float(self._inner, content)
        else:
            return super().is_right(content)

    def is_over_or_right(self, content: Union[FloatSet, float]) -> bool:
        """
        Returns whether ``self`` is to the right of ``other`` allowing overlap.
        That is, ``self`` ends after ``other`` ends (or at the same value).

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            overright_set_set, overright_floatset_float
        """
        if isinstance(content, float):
            return overright_floatset_float(self._inner, content)
        else:
            return super().is_over_or_right(content)

    # ------------------------- Set Operations --------------------------------

    @overload
    def intersection(self, other: float) -> Optional[float]:
        ...

    @overload
    def intersection(self, other: FloatSet) -> Optional[FloatSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: A :class:`FloatSet` or :class:`float` instance

        Returns:
            An object of the same type as ``other`` or ``None`` if the intersection is empty.

        MEOS Functions:
            intersection_set_set, intersection_floatset_float
        """
        if isinstance(other, float):
            return intersection_floatset_float(self._inner, other)
        elif isinstance(other, FloatSet):
            result = intersection_set_set(self._inner, other._inner)
            return FloatSet(_inner=result) if result is not None else None
        else:
            return super().intersection(other)

    def minus(self, other: Union[FloatSet, float]) -> Optional[FloatSet]:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: A :class:`FloatSet` or :class:`float` instance

        Returns:
            A :class:`FloatSet` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_set_set, minus_floatset_float
        """
        if isinstance(other, float):
            result = minus_floatset_float(self._inner, other)
            return FloatSet(_inner=result) if result is not None else None
        elif isinstance(other, FloatSet):
            result = minus_set_set(self._inner, other._inner)
            return FloatSet(_inner=result) if result is not None else None
        else:
            return super().minus(other)

    def subtract_from(self, other: float) -> Optional[float]:
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: A :class:`float` instance

        Returns:
            A :class:`float` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_float_floatset

        See Also:
            :meth:`minus`
        """
        return minus_float_floatset(other, self._inner)

    def union(self, other: Union[FloatSet, float]) -> FloatSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: A :class:`FloatSet` or :class:`float` instance

        Returns:
            A :class:`FloatSet` instance.

        MEOS Functions:
            union_set_set, union_floatset_float
        """
        if isinstance(other, float):
            result = union_floatset_float(self._inner, other)
            return FloatSet(_inner=result) if result is not None else None
        elif isinstance(other, FloatSet):
            result = union_set_set(self._inner, other._inner)
            return FloatSet(_inner=result) if result is not None else None
        else:
            return super().union(other)

    # ------------------------- Distance Operations ---------------------------

    def distance(self, other: Union[float, FloatSet, FloatSpan, FloatSpanSet]) -> float:
        if isinstance(other, float):
            return distance_floatset_float(self._inner, other)
        else:
            return super().distance(other)



from datetime import datetime
from typing import Union

from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..collections import TimestampSet, Period, PeriodSet


class TimeInstantaneousUnionAggregator(BaseAggregator[Union[datetime, TimestampSet], TimestampSet]):
    """
    Temporal union of instantaneous time objects (:class:'~datetime.datetime' and
    :class:`~pymeos.time.timestampset.TimestampSet`).

    MEOS Functions:
        timestamp_union_transfn, set_union_transfn, set_union_finalfn
    """

    @classmethod
    def _add(cls, state, temporal):
        if isinstance(temporal, datetime):
            state = timestamp_union_transfn(state, datetime_to_timestamptz(temporal))
        elif isinstance(temporal, TimestampSet):
            state = set_union_transfn(state, temporal._inner)
        else:
            cls._error(temporal)
        return state

    @classmethod
    def _finish(cls, state) -> TimestampSet:
        result = set_union_finalfn(state)
        return TimestampSet(_inner=result)


class TimeContinuousUnionAggregator(BaseAggregator[Union[Period, PeriodSet], PeriodSet]):
    """
    Temporal union of continuous time objects (:class:`~pymeos.time.period.Period` and
    :class:`~pymeos.time.periodset.PeriodSet`).

    MEOS Functions:
        union_spanset_span, union_spanset_spanset
    """

    @classmethod
    def _add(cls, state, temporal):
        if isinstance(temporal, Period):
            state = union_spanset_span(state, temporal._inner)
        elif isinstance(temporal, PeriodSet):
            state = union_spanset_spanset(state, temporal._inner)
        else:
            cls._error(temporal)
        return state

    @classmethod
    def _finish(cls, state) -> PeriodSet:
        return PeriodSet(_inner=state)



from datetime import datetime
from typing import Union

from .period import Period
from .periodset import PeriodSet
from .timestampset import TimestampSet

Time = Union[datetime, TimestampSet, Period, PeriodSet]
"""
Union type that includes all Time types in PyMEOS:  

- :class:`~datetime.datetime` for timestamps  
- :class:`~pymeos.time.timestampset.TimestampSet` for sets of timestamps  
- :class:`~pymeos.time.period.Period` for periods of time  
- :class:`~pymeos.time.periodset.PeriodSet` for sets of periods of time  
"""



from __future__ import annotations

from enum import IntEnum


class TInterpolation(IntEnum):
    """
    Enum for representing the different types of interpolation present in
    PyMEOS.
    """
    NONE = 0
    DISCRETE = 1
    STEPWISE = 2
    LINEAR = 3

    @staticmethod
    def from_string(source: str, none: bool = True) -> TInterpolation:
        """
        Returns the :class:`TInterpolation` element equivalent to `source`.

        Args:
            source: :class:`string` representing the interpolation
            none: indicates whether to return `TIntepolation.NONE` when
            `source` represents an invalid interpolation

        Returns:
            A new :class:`TInterpolation` instance.

        Raises:
            ValueError: when `source` doesn't represent any valid interpolation
            and `none` is False

        """
        if source.lower() == 'discrete':
            return TInterpolation.DISCRETE
        elif source.lower() == 'linear':
            return TInterpolation.LINEAR
        elif source.lower() == 'stepwise' or source.lower() == 'step':
            return TInterpolation.STEPWISE
        elif source.lower() == 'none' or none:
            return TInterpolation.NONE
        else:
            raise ValueError(f"Value {source} doesn't represent any valid interpolation")



from __future__ import annotations

from abc import ABC
from typing import Optional, List, Union, Any, TypeVar, Type

from pandas import DataFrame
from pymeos_cffi import *

from ..temporal.temporal import Temporal

TBase = TypeVar('TBase')
TG = TypeVar('TG', bound='Temporal[Any]')
TI = TypeVar('TI', bound='TInstant[Any]')
TS = TypeVar('TS', bound='TSequence[Any]')
TSS = TypeVar('TSS', bound='TSequenceSet[Any]')
Self = TypeVar('Self', bound='TSequenceSet[Any]')


class TSequenceSet(Temporal[TBase, TG, TI, TS, TSS], ABC):
    """
    Base class for temporal sequence set types, i.e. temporal values that are
    defined by a set of temporal sequences.
    """

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, Any]]] = None,
                 normalize: bool = True, _inner=None):
        assert (_inner is not None) or ((string is not None) != (sequence_list is not None)), \
            "Either string must be not None or sequence_list must be not"
        if _inner is not None:
            self._inner = as_tsequenceset(_inner)
        elif string is not None:
            self._inner = as_tsequenceset(self.__class__._parse_function(string))
        else:
            sequences = [x._inner if isinstance(x, self.ComponentClass) \
                             else self.__class__._parse_function(x) for x in sequence_list]
            count = len(sequences)
            self._inner = tsequenceset_make(sequences, count, normalize)

    @classmethod
    def from_sequences(cls: Type[Self],
                       sequence_list: Optional[List[Union[str, Any]]] = None,
                       normalize: bool = True) -> Self:
        """
        Create a temporal sequence set from a list of sequences.

        Args:
            sequence_list: List of sequences.
            normalize: Whether to normalize the temporal sequence set.

        Returns:
            A temporal sequence set.
        """
        return cls(sequence_list=sequence_list, normalize=normalize)

    # ------------------------- Accessors -------------------------------------
    def num_sequences(self) -> int:
        """
        Returns the number of sequences in ``self``.
        """
        return temporal_num_sequences(self._inner)

    def start_sequence(self) -> TS:
        """
        Returns the first sequence in ``self``.
        """
        return self.ComponentClass(_inner=temporal_start_sequence(self._inner))

    def end_sequence(self) -> TS:
        """
        Returns the last sequence in ``self``.
        """
        return self.ComponentClass(_inner=temporal_end_sequence(self._inner))

    def sequence_n(self, n) -> TS:
        """
        Returns the ``n``-th sequence in ``self``.
        """
        return self.ComponentClass(_inner=temporal_sequence_n(self._inner, n + 1))

    def sequences(self) -> List[TS]:
        """
        Returns the list of sequences in ``self``.
        """
        ss, count = temporal_sequences(self._inner)
        return [self.ComponentClass(_inner=ss[i]) for i in range(count)]

    # ------------------------- Transformations -------------------------------
    def to_dataframe(self) -> DataFrame:
        """
        Returns a pandas DataFrame representation of ``self``.
        """
        sequences = self.sequences()
        data = {
            'sequence': [i for i, seq in enumerate(sequences) \
                         for _ in range(seq.num_instants())],
            'time': [t for seq in sequences for t in seq.timestamps()],
            'value': [v for seq in sequences for v in seq.values()]
        }
        return DataFrame(data).set_index(keys=['sequence', 'time'])

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        """
        Plot the temporal sequence set.

        See Also:
            :meth:`pymeos.plotters.TemporalSequenceSetPlotter.plot`
        """
        from ..plotters import TemporalSequenceSetPlotter
        return TemporalSequenceSetPlotter.plot(self, *args, **kwargs)



from __future__ import annotations

from typing import Optional, overload, Union

from pymeos_cffi import *

from ..base import Set


class TextSet(Set[str]):
    """
    Class for representing a set of text values.

    ``TextSet`` objects can be created with a single argument of type string as
    in MobilityDB.

        >>> TextSet(string='{a, b, c, def}')

    Another possibility is to create a ``TextSet`` object from a list of strings.

        >>> TextSet(elements=['a', 'b', 'c', 'def'])


    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'textset'

    _parse_function = textset_in
    _parse_value_function = lambda x: x
    _make_function = textset_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------

    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            textset_out
        """
        return textset_out(self._inner)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self):
        raise NotImplementedError()

    def to_span(self):
        raise NotImplementedError()

    # ------------------------- Accessors -------------------------------------

    def start_element(self):
        """
        Returns the first element in ``self``.

        Returns:
            A :class:`str` instance

        MEOS Functions:
            textset_start_value
        """
        return textset_start_value(self._inner)

    def end_element(self):
        """
        Returns the last element in ``self``.

        Returns:
            A :class:`str` instance

        MEOS Functions:
            textset_end_value
        """
        return textset_end_value(self._inner)

    def element_n(self, n: int):
        """
        Returns the ``n``-th element in ``self``.

        Args:
            n: The 0-based index of the element to return.

        Returns:
            A :class:`str` instance

        MEOS Functions:
            textset_value_n
        """
        super().element_n(n)
        return text2cstring(textset_value_n(self._inner, n + 1)[0])

    def elements(self):
        """
        Returns the elements in ``self``.

        Returns:
            A list of :class:`str` instances

        MEOS Functions:
            textset_values
        """
        elems = textset_values(self._inner)
        return [text2cstring(elems[i]) for i in range(self.num_elements())]

    # ------------------------- Topological Operations --------------------------------

    def contains(self, content: Union[TextSet, str]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set, contains_textset_text
        """
        if isinstance(content, str):
            return contains_textset_text(self._inner, content)
        else:
            return super().contains(content)


    # ------------------------- Transformations --------------------------------
    def lowercase(self):
        """
        Returns a new textset that is the result of appling uppercase to ``self``

        Returns:
            A :class:`str` instance

        MEOS Functions:
            textset_lower
        """
        return self.__class__(_inner=textset_lower(self._inner))

    def uppercase(self):
        """
        Returns a new textset that is the result of appling uppercase to ``self``

        Returns:
            A :class:`str` instance

        MEOS Functions:
            textset_upper
        """
        return self.__class__(_inner=textset_upper(self._inner))

    # ------------------------- Set Operations --------------------------------

    @overload
    def intersection(self, other: str) -> Optional[str]:
        ...

    @overload
    def intersection(self, other: TextSet) -> Optional[TextSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: A :class:`TextSet` or :class:`str` instance

        Returns:
            An object of the same type as ``other`` or ``None`` if the intersection is empty.

        MEOS Functions:
            intersection_textset_text, intersection_set_set
        """
        if isinstance(other, str):
            result = intersection_textset_text(self._inner, other)
            return text2cstring(result[0]) if result is not None else None
        elif isinstance(other, TextSet):
            result = intersection_set_set(self._inner, other._inner)
            return TextSet(_inner=result) if result is not None else None
        else:
            return super().intersection(other)

    def minus(self, other: Union[TextSet, str]) -> Optional[TextSet]:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: A :class:`TextSet` or :class:`str` instance

        Returns:
            A :class:`TextSet` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_textset_text, minus_set_set
        """
        if isinstance(other, str):
            result = minus_textset_text(self._inner, other)
            return TextSet(_inner=result) if result is not None else None
        elif isinstance(other, TextSet):
            result = minus_set_set(self._inner, other._inner)
            return TextSet(_inner=result) if result is not None else None
        else:
            return super().minus(other)

    def subtract_from(self, other: str) -> Optional[str]:
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: A :class:`str` instance

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            minus_geo_geoset

        See Also:
            :meth:`minus`
        """
        result = minus_text_textset(other, self._inner)
        return text2cstring(result[0]) if result is not None else None

    def union(self, other: Union[TextSet, str]) -> TextSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: A :class:`TextSet` or :class:`str` instance

        Returns:
            A :class:`TextSet` instance.

        MEOS Functions:
            union_textset_text, union_set_set
        """
        if isinstance(other, str):
            result = union_textset_text(self._inner, other)
            return TextSet(_inner=result) if result is not None else None
        elif isinstance(other, TextSet):
            result = union_set_set(self._inner, other._inner)
            return TextSet(_inner=result) if result is not None else None
        else:
            return super().union(other)



from abc import ABC
from datetime import datetime

from ..base.collection import Collection


class TimeCollection(Collection[datetime], ABC):

    def is_before(self, other) -> bool:
        return self.is_left(other)

    def is_over_or_before(self, other) -> bool:
        return self.is_over_or_left(other)

    def is_over_or_after(self, other) -> bool:
        return self.is_over_or_right(other)

    def is_after(self, other) -> bool:
        return self.is_right(other)



from typing import Union

from matplotlib import pyplot as plt
from pymeos import IntSpan, FloatSpan


class SpanPlotter:
    """
    Plotter for :class:`FloatSpan` and :class:`IntSpan` objects.
    """

    @staticmethod
    def plot_span(span: Union[IntSpan, FloatSpan], *args, axes=None, **kwargs):
        """
        Plot a :class:`FloatSpan` or :class:`IntSpan` on the given axes.

        Params:
            span: The :class:`FloatSpan` or :class:`IntSpan` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        ll = base.axhline(span.lower(), *args, linestyle='-' if span.lower_inc() else '--', **kwargs)
        kwargs.pop('label', None)
        ul = base.axhline(span.upper(), *args, linestyle='-' if span.upper_inc() else '--', **kwargs)
        s = base.axhspan(span.lower(), span.upper(), *args, alpha=0.3, **kwargs)
        return [ll, ul, s]



from __future__ import annotations

from typing import Union, overload, Optional, TYPE_CHECKING, List

from pymeos_cffi import floatspanset_in, floatspanset_out, spanset_width, \
    floatspanset_shift_scale, adjacent_floatspanset_float, contains_floatspanset_float, \
    spanset_eq, float_to_floatspanset, left_floatspanset_float, \
    overleft_floatspanset_float, right_floatspanset_float, overright_floatspanset_float, \
    distance_floatspanset_float, intersection_floatspanset_float, \
    union_floatspanset_float, minus_floatspanset_float, floatspanset_intspanset

from pymeos.collections import SpanSet

if TYPE_CHECKING:
    from .floatspan import FloatSpan
    from .intspanset import IntSpanSet


class FloatSpanSet(SpanSet[float]):
    """
    Class for representing lists of disjoint floatspans.

    ``FloatSpanSet`` objects can be created with a single argument of type string
    as in MobilityDB.

        >>> FloatSpanSet(string='{[8, 10], [11, 1]}')

    Another possibility is to give a list specifying the composing
    spans, which can be instances  of ``str`` or ``FloatSpan``. The composing
    spans must be given in increasing order.

        >>> FloatSpanSet(span_list=['[8, 10]', '[11, 12]'])
        >>> FloatSpanSet(span_list=[FloatSpan('[8, 10]'), FloatSpan('[11, 12]')])

    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'floatspanset'

    _parse_function = floatspanset_in
    _parse_value_function = lambda span: floatspanset_in(span)[0] if isinstance(span, str) else span._inner[0]

    # ------------------------- Output ----------------------------------------
    def __str__(self, max_decimals: int = 15):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            floatspanset_out
        """
        return floatspanset_out(self._inner, max_decimals)

    # ------------------------- Conversions -----------------------------------

    def to_span(self) -> FloatSpan:
        """
        Returns a span that encompasses ``self``.

        Returns:
            A new :class:`FloatSpan` instance

        MEOS Functions:
            spanset_span
        """
        from .floatspan import FloatSpan
        return FloatSpan(_inner=super().to_span())

    def to_intspanset(self) -> IntSpanSet:
        """
        Returns an intspanset that encompasses ``self``.

        Returns:
            A new :class:`IntSpanSet` instance

        MEOS Functions:
            floatspanset_intspanset
        """
        from .intspanset import IntSpanSet
        return IntSpanSet(_inner=floatspanset_intspanset(self._inner))

    # ------------------------- Accessors -------------------------------------

    def width(self, ignore_gaps: Optional[bool] = False) -> float:
        """
        Returns the width of the spanset. By default, i.e., when the second
        argument is False, the function takes into account the gaps within,
        i.e., returns the sum of the widths of the spans within.
        Otherwise, the function returns the width of the spanset ignoring
        any gap, i.e., the width from the lower bound of the first span to
        the upper bound of the last span.

        Parameters:
            ignore_gaps: Whether to take into account potential gaps in
            the spanset.

        Returns:
            A `float` representing the duration of the spanset

        MEOS Functions:
            spanset_width
        """
        return spanset_width(self._inner, ignore_gaps)

    def start_span(self) -> FloatSpan:
        """
        Returns the first span in ``self``.
        Returns:
            A :class:`FloatSpan` instance

        MEOS Functions:
            spanset_start_span
        """
        from .floatspan import FloatSpan
        return FloatSpan(_inner=super().start_span())

    def end_span(self) -> FloatSpan:
        """
        Returns the last span in ``self``.
        Returns:
            A :class:`FloatSpan` instance

        MEOS Functions:
            spanset_end_span
        """
        from .floatspan import FloatSpan
        return FloatSpan(_inner=super().end_span())

    def span_n(self, n: int) -> FloatSpan:
        """
        Returns the n-th span in ``self``.
        Returns:
            A :class:`FloatSpan` instance

        MEOS Functions:
            spanset_span_n
        """
        from .floatspan import FloatSpan
        return FloatSpan(_inner=super().span_n(n))

    def spans(self) -> List[FloatSpan]:
        """
        Returns the list of spans in ``self``.
        Returns:
            A :class:`list[FloatSpan]` instance

        MEOS Functions:
            spanset_spans
        """
        from .floatspan import FloatSpan
        ps = super().spans()
        return [FloatSpan(_inner=ps[i]) for i in range(self.num_spans())]

    # ------------------------- Transformations -------------------------------

    def shift(self, delta: int) -> FloatSpanSet:
        """
        Return a new ``FloatSpanSet`` with the lower and upper bounds shifted by
        ``delta``.

        Args:
            delta: The value to shift by

        Returns:
            A new ``FloatSpanSet`` instance

        MEOS Functions:
            floatspanset_shift_scale
        """
        return self.shift_scale(delta, None)

    def scale(self, width: int) -> FloatSpanSet:
        """
        Return a new ``FloatSpanSet`` with the lower and upper bounds scaled so
        that the width is ``width``.

        Args:
            width: The new width

        Returns:
            A new ``FloatSpanSet`` instance

        MEOS Functions:
            floatspanset_shift_scale
        """
        return self.shift_scale(None, width)

    def shift_scale(self, delta: Optional[int], width: Optional[int]) -> FloatSpanSet:
        """
        Return a new ``FloatSpanSet`` with the lower and upper bounds shifted by
        ``delta`` and scaled so that the width is ``width``.

        Args:
            delta: The value to shift by
            width: The new width

        Returns:
            A new ``FloatSpanSet`` instance

        MEOS Functions:
            floatspanset_shift_scale
        """
        d = delta if delta is not None else 0
        w = width if width is not None else 0
        modified = floatspanset_shift_scale(self._inner, d, w, delta is not None,
                                            width is not None)
        return FloatSpanSet(_inner=modified)

    # ------------------------- Topological Operations --------------------------------

    def is_adjacent(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share
        a bound but only one of them contains it.

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_spanset_span, adjacent_spanset_spanset,
            adjacent_floatspanset_float
        """
        if isinstance(other, int):
            return adjacent_floatspanset_float(self._inner, other)
        else:
            return super().is_adjacent(other)

    def contains(self, content: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_spanset_span, contains_spanset_spanset,
            contains_floatspanset_float
        """
        if isinstance(content, int):
            return contains_floatspanset_float(self._inner, content)
        else:
            return super().contains(content)

    def is_same(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` and the bounding period of ``other`` is the
        same.

        Args:
            other: object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            same_period_temporal
        """
        if isinstance(other, int):
            return spanset_eq(self._inner, float_to_floatspanset(other))
        else:
            return super().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly left of ``other``. That is, 
        ``self`` ends before ``other`` starts.

        Args:
            other: object to compare with

        Returns:
            True if left, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset, left_floatspan_float
        """
        if isinstance(other, int):
            return left_floatspanset_float(self._inner, other)
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is left ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset, overleft_floatspan_float
        """
        if isinstance(other, int):
            return overleft_floatspanset_float(self._inner, other)
        else:
            return super().is_over_or_left(other)

    def is_right(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly right ``other``. That is, ``self``
        starts after ``other`` ends.

        Args:
            other: object to compare with

        Returns:
            True if right, False otherwise

        MEOS Functions:
            right_span_span, right_span_spanset, right_floatspan_float
        """
        if isinstance(other, int):
            return right_floatspanset_float(self._inner, other)
        else:
            return super().is_right(other)

    def is_over_or_right(self, other: Union[int, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is right ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_spanset_span, overright_spanset_spanset,
            overright_floatspanset_float
        """
        if isinstance(other, int):
            return overright_floatspanset_float(self._inner, other)
        else:
            return super().is_over_or_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[int, FloatSpan, FloatSpanSet]) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other: object to compare with

        Returns:
            A float value

        MEOS Functions:
            distance_spanset_span, distance_spanset_spanset,
            distance_floatspanset_float
        """
        if isinstance(other, int):
            return distance_floatspanset_float(self._inner, other)
        else:
            return super().distance(other)

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: Union[int, float]) -> Optional[float]:
        ...

    @overload
    def intersection(self, other: FloatSpan) -> Optional[FloatSpanSet]:
        ...

    @overload
    def intersection(self, other: FloatSpanSet) -> Optional[FloatSpanSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            An int or a :class:`FloatSpanSet` instance. The actual class depends
            on ``other``.

        MEOS Functions:
            intersection_floatspanset_float, intersection_spanset_spanset,
            intersection_spanset_span
        """
        if isinstance(other, int) or isinstance(other, float):
            result = intersection_floatspanset_float(self._inner, float(other))
        else:
            result = super().intersection(other)
        return FloatSpanSet(_inner=result) if result is not None else None

    def __mul__(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_floatspanset_float, intersection_spanset_spanset,
            intersection_spanset_span
        """
        return self.intersection(other)

    def minus(self, other: Union[int, FloatSpan, FloatSpanSet]) -> FloatSpanSet:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`FloatSpanSet` instance.

        MEOS Functions:
            minus_spanset_span, minus_spanset_spanset, minus_floatspanset_float
        """
        if isinstance(other, int) or isinstance(other, float):
            result = minus_floatspanset_float(self._inner, float(other))
        else:
            result = super().minus(other)
        return FloatSpanSet(_inner=result) if result is not None else None

    def __sub__(self, other):
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`FloatSpanSet` instance.

        MEOS Functions:
            minus_spanset_span, minus_spanset_spanset,
            minus_floatspanset_float
        """
        return self.minus(other)

    def union(self, other: Union[int, FloatSpan, FloatSpanSet]) -> FloatSpanSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`FloatSpanSet` instance.

        MEOS Functions:
            union_floatspanset_float, union_spanset_spanset,
            union_spanset_span
        """
        if isinstance(other, int) or isinstance(other, float):
            result = union_floatspanset_float(self._inner, float(other))
        else:
            result = super().union(other)
        return FloatSpanSet(_inner=result) if result is not None else None

    def __add__(self, other):
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`FloatSpanSet` instance.

        MEOS Functions:
            union_floatspanset_float, union_spanset_spanset,
            union_spanset_span
        """
        return self.union(other)



from functools import partial
from typing import Union, List

import matplotlib.pyplot as plt

from ..temporal import TSequence, TInterpolation, TInstant


class TemporalSequencePlotter:
    """
    Plotter for :class:`TSequence` and lists of :class:`TInstant`.
    """

    @staticmethod
    def plot(sequence: Union[TSequence, List[TInstant]], *args, axes=None, show_markers=True, show_grid=True, **kwargs):
        """
        Plot a :class:`TSequence` or a list of :class:`TInstant` on the given axes. The actual plot function is chosen
        based on the interpolation of the sequence.

        Params:
            sequence: The :class:`TSequence` or list of :class:`TInstant` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            show_markers: Whether to show markers at the start and end of the sequence. The marker will be filled if the
            sequence is inclusive at that end, and empty otherwise.
            show_grid: Whether to show a grid.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        if isinstance(sequence, list):
            plot_func = base.scatter
            show_markers = False
        elif sequence.interpolation() == TInterpolation.LINEAR:
            plot_func = base.plot
        elif sequence.interpolation() == TInterpolation.STEPWISE:
            plot_func = partial(base.step, where='post')
        else:
            plot_func = base.scatter
            show_markers = False

        ins = sequence.instants() if isinstance(sequence, TSequence) else sequence
        x = [i.timestamp() for i in ins]
        y = [i.value() for i in ins]

        base.set_axisbelow(True)

        if show_grid:
            base.grid(zorder=0.5)
        plots = [plot_func(x, y, *args, **kwargs)]

        if show_markers:
            color = plots[0][0].get_color()
            plots.append(base.scatter(x[0], y[0], s=40, marker='o',
                                      facecolors=color if sequence.lower_inc() else 'none',
                                      edgecolors=color, zorder=2 if sequence.lower_inc() else 3))
            plots.append(base.scatter(x[-1], y[-1], s=40, marker='o',
                                      facecolors=color if sequence.upper_inc() else 'none',
                                      edgecolors=color, zorder=2 if sequence.upper_inc() else 3))

        if isinstance(y[0], bool):
            plt.yticks([1.0, 0.0], ['True', 'False'])
            plt.ylim(-0.25, 1.25)

        base.tick_params(axis="x", rotation=45)

        return plots

    @staticmethod
    def plot_sequences(sequences: List[TSequence], *args, **kwargs):
        """
        Plot a list of :class:`TSequence` on the given axes. Every sequence will be plotted in a different color.

        Params:
            sequences: The list of :class:`TSequence` to plot.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List of lists with the plotted elements of each sequence.

        See Also:
            :func:`~pymeos.plotters.sequence_plotter.TemporalSequencePlotter.plot`,
            :meth:`~pymeos.plotters.sequenceset_plotter.TemporalSequenceSetPlotter.plot`
        """
        plots = []
        for seq in sequences:
            plots.append(TemporalSequencePlotter.plot(seq, *args, **kwargs))
        return plots



from typing import Optional

from pymeos_cffi import meos_initialize, meos_finalize


def pymeos_initialize(timezone: Optional[str] = None) -> None:
    """
    Initializes the underlying MEOS platform.
    Must be called before any other PyMEOS function.

    Args:
        timezone: :class:`str` indicating the desired timezone to be used. Defaults to system timezone.

    MEOS Functions:
        meos_initialize
    """
    meos_initialize(timezone)


def pymeos_finalize() -> None:
    """
    Cleans up the underlying MEOS platform.
    Should be called at the end after PyMEOS is no longer used.

    MEOS Functions:
        meos_finalize
    """
    meos_finalize()



from typing import List, Union

from .sequence_plotter import TemporalSequencePlotter
from .. import TSequence
from ..temporal import TSequenceSet


class TemporalSequenceSetPlotter:
    """
    Plotter for :class:`TSequenceSet` and lists of :class:`TSequence`.
    """

    @staticmethod
    def plot(sequence_set: Union[TSequenceSet, List[TSequence]], *args, **kwargs):
        """
        Plot a :class:`TSequenceSet` or a list of :class:`TSequence` on the given axes. Every sequence in the set will be
        plotted with the same color.

        Params:
            sequence_set: The :class:`TSequenceSet` or list of :class:`TSequence` to plot.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.

        See Also:
            :func:`~pymeos.plotters.sequence_plotter.TemporalSequencePlotter.plot`

        """
        seqs = sequence_set.sequences() if isinstance(sequence_set, TSequenceSet) else sequence_set
        plots = [TemporalSequencePlotter.plot(seqs[0], *args, **kwargs)]
        if 'color' not in kwargs:
            kwargs['color'] = plots[0][0].get_color()
        kwargs.pop('label', None)
        for seq in seqs[1:]:
            plots.append(TemporalSequencePlotter.plot(seq, *args, **kwargs))
        return plots



from datetime import datetime
from typing import Union

from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..boxes import Box
from ..main import TIntSeq, TIntSeqSet
from ..temporal import Temporal, TInterpolation
from ..collections import Time, TimestampSet, Period, PeriodSet


class TemporalInstantCountAggregator(BaseAggregator[
                                         Union[datetime, TimestampSet, Temporal],
                                         TIntSeq]):
    """
    Temporal count for instantaneous temporal objects:

     - :class:`~datetime.datetime`
     - :class:`~pymeos.time.timestampset.TimestampSet`
     - :class:`~pymeos.temporal.temporal.Temporal` with Discrete interpolation

    MEOS Functions:
        timestamp_tcount_transfn, timestampset_tcount_transfn, temporal_tcount_transfn, temporal_tagg_finalfn
    """

    @classmethod
    def _add(cls, state, temporal):
        if isinstance(temporal, datetime):
            state = timestamp_tcount_transfn(state, datetime_to_timestamptz(temporal))
        elif isinstance(temporal, TimestampSet):
            state = timestampset_tcount_transfn(state, temporal._inner)
        elif isinstance(temporal, Temporal) and temporal.interpolation() == TInterpolation.DISCRETE:
            state = temporal_tcount_transfn(state, temporal._inner)
        else:
            cls._error(temporal)
        return state


class TemporalPeriodCountAggregator(BaseAggregator[Union[Period, PeriodSet, Temporal], TIntSeqSet]):
    """
    Temporal count for non-instantaneous temporal objects:

     - :class:`~pymeos.time.period.Period`
     - :class:`~pymeos.time.periodset.PeriodSet`
     - :class:`~pymeos.temporal.temporal.Temporal` without Discrete interpolation

    MEOS Functions:
        period_tcount_transfn, periodset_tcount_transfn, temporal_tcount_transfn, temporal_tagg_finalfn
    """

    @classmethod
    def _add(cls, state, temporal):
        if isinstance(temporal, Period):
            state = period_tcount_transfn(state, temporal._inner)
        elif isinstance(temporal, PeriodSet):
            state = periodset_tcount_transfn(state, temporal._inner)
        elif isinstance(temporal, Temporal) and temporal.interpolation() != TInterpolation.DISCRETE:
            state = temporal_tcount_transfn(state, temporal._inner)
        else:
            cls._error(temporal)
        return state


class TemporalExtentAggregator(BaseAggregator[Union[Time, Temporal], Period]):
    """
    Temporal extent of any kind of temporal object, i.e. smallest :class:`~pymeos.time.period.Period` that includes
    all aggregated temporal objects.

    MEOS Functions:
        temporal_extent_transfn, timestamp_extent_transfn, timestampset_extent_transfn, span_extent_transfn,
         spanset_extent_transfn, temporal_tagg_finalfn
    """

    @classmethod
    def _add(cls, state, temporal):
        if isinstance(temporal, Temporal):
            state = temporal_extent_transfn(state, temporal._inner)
        elif isinstance(temporal, datetime):
            state = timestamp_extent_transfn(state, datetime_to_timestamptz(temporal))
        elif isinstance(temporal, TimestampSet):
            state = set_extent_transfn(state, temporal._inner)
        elif isinstance(temporal, Period):
            state = span_extent_transfn(state, temporal._inner)
            pass
        elif isinstance(temporal, PeriodSet):
            state = spanset_extent_transfn(state, temporal._inner)
        else:
            cls._error(temporal)
        return state

    @classmethod
    def _finish(cls, state) -> Union[Temporal, Time, Box]:
        return Period(_inner=state)



from .base import *
from .number import *
from .text import *
from .time import *
from .text import *
from .geo import *

__all__ = [
    'Set', 'Span', 'SpanSet',
    'Time', 'TimestampSet', 'Period', 'PeriodSet', 'datetime', 'timedelta',
    'TextSet',
    'GeoSet', 'GeometrySet', 'GeographySet',
    'IntSet', 'IntSpan', 'IntSpanSet',
    'FloatSet', 'FloatSpan', 'FloatSpanSet'
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union
from typing import TypeVar, Type, Callable, Any, TYPE_CHECKING, List

from pymeos_cffi import *

from .collection import Collection

if TYPE_CHECKING:
    from .span import Span

T = TypeVar('T')
Self = TypeVar('Self', bound='Span[Any]')


class SpanSet(Collection[T], ABC):
    """
    Base class for all spanset classes.
    """

    __slots__ = ['_inner']

    _parse_function: Callable[[str], 'CData'] = None
    _parse_value_function: Callable[[Union[str, T]], Any] = None

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *, span_list: Optional[List[Union[str, Span]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__()
        assert (_inner is not None) or ((string is not None) != (span_list is not None)), \
            "Either string must be not None or span_list must be not"
        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = self.__class__._parse_function(string)
        else:
            spans = [self.__class__._parse_value_function(p) for p in span_list]
            self._inner = spanset_make(spans, normalize)

    def __copy__(self: Self) -> Self:
        """
        Return a copy of ``self``.

        Returns:
            A new :class:`SpanSet` instance

        MEOS Functions:
            spanset_copy
        """
        inner_copy = spanset_copy(self._inner)
        return self.__class__(_inner=inner_copy)

    @classmethod
    def from_wkb(cls: Type[Self], wkb: bytes) -> Self:
        """
        Returns a `SpanSet` from its WKB representation.

        Args:
            wkb: The WKB string.

        Returns:
            A new :class:`SpanSet` instance

        MEOS Functions:
            spanset_from_wkb
        """
        result = spanset_from_wkb(wkb)
        return cls(_inner=result)

    @classmethod
    def from_hexwkb(cls: Type[Self], hexwkb: str) -> Self:
        """
        Returns a `SpanSet` from its WKB representation in hex-encoded ASCII.
        Args:
            hexwkb: WKB representation in hex-encoded ASCII

        Returns:
            A new :class:`SpanSet` instance

        MEOS Functions:
            spanset_from_hexwkb
        """
        result = spanset_from_hexwkb(hexwkb)
        return cls(_inner=result)

    # ------------------------- Output ----------------------------------------
    @abstractmethod
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance
        """
        raise NotImplementedError()

    def __repr__(self):
        """
        Return the string representation of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            periodset_out
        """
        return (f'{self.__class__.__name__}'
                f'({self})')

    def as_wkb(self) -> bytes:
        """
        Returns the WKB representation of ``self``.

        Returns:
            A :class:`str` object with the WKB representation of ``self``.

        MEOS Functions:
            spanset_as_wkb
        """
        return spanset_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the WKB representation of ``self`` in hex-encoded ASCII.
        Returns:
            A :class:`str` object with the WKB representation of ``self`` in hex-encoded ASCII.

        MEOS Functions:
            spanset_as_hexwkb
        """
        return spanset_as_hexwkb(self._inner, -1)[0]

    # ------------------------- Conversions -----------------------------------
    @abstractmethod
    def to_span(self) -> Span:
        """
        Returns a span that encompasses ``self``.

        Returns:
            A new :class:`Span` instance

        MEOS Functions:
            spanset_span
        """
        return spanset_span(self._inner)

    # ------------------------- Accessors -------------------------------------

    def num_spans(self) -> int:
        """
        Returns the number of spans in ``self``.
        Returns:
            An :class:`int`

        MEOS Functions:
            spanset_num_spans
        """
        return spanset_num_spans(self._inner)

    @abstractmethod
    def start_span(self) -> Span:
        """
        Returns the first span in ``self``.
        Returns:
            A :class:`Span` instance

        MEOS Functions:
            spanset_start_span
        """
        return spanset_start_span(self._inner)

    @abstractmethod
    def end_span(self) -> Span:
        """
        Returns the last span in ``self``.
        Returns:
            A :class:`Span` instance

        MEOS Functions:
            spanset_end_span
        """
        return spanset_end_span(self._inner)

    @abstractmethod
    def span_n(self, n: int) -> Span:
        """
        Returns the n-th span in ``self``.
        Returns:
            A :class:`Span` instance

        MEOS Functions:
            spanset_span_n
        """
        return spanset_span_n(self._inner, n + 1)

    @abstractmethod
    def spans(self) -> List[Span]:
        """
        Returns the list of periods in ``self``.
        Returns:
            A :class:`list[Period]` instance

        MEOS Functions:
            spanset_spans
        """
        return spanset_spans(self._inner)

    def __hash__(self) -> int:
        """
        Return the hash representation of ``self``.

        Returns:
            A new :class:`int` instance

        MEOS Functions:
            spanset_hash
        """
        return spanset_hash(self._inner)

    # ------------------------- Transformations -------------------------------

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share a bound but only one of them
        contains it.

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_spanset_span, adjacent_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return adjacent_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return adjacent_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_contained_in(self, container) -> bool:
        """
        Returns whether ``self`` is contained in ``container``.

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_spanset_span, contained_spanset_spanset
        """
        from .span import Span
        if isinstance(container, Span):
            return contained_spanset_span(self._inner, container._inner)
        elif isinstance(container, SpanSet):
            return contained_spanset_spanset(self._inner, container._inner)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')

    def contains(self, content) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
        contains_spanset_span, contains_spanset_spanset,
        """
        from .span import Span
        if isinstance(content, Span):
            return contains_spanset_span(self._inner, content._inner)
        elif isinstance(content, SpanSet):
            return contains_spanset_spanset(self._inner, content._inner)
        else:
            raise TypeError(f'Operation not supported with type {content.__class__}')

    def __contains__(self, item):
        """
        Returns whether ``self`` contains ``content``.

        Args:
            item: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_spanset_span, contains_spanset_spanset
        """
        return self.contains(item)

    def overlaps(self, other) -> bool:
        """
        Returns whether ``self`` overlaps ``other``. That is, both share at least an instant

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_spanset_span, overlaps_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return overlaps_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overlaps_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_same(self, other) -> bool:
        """
        Returns whether the bounding span of `self` is the same as the bounding span of `other`.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if same, False otherwise.

        See Also:
            :meth:`Period.is_same`
        """
        return self.to_span().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``. That is, ``self`` ends before ``other`` starts.

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
        before_periodset_timestamp, left_spanset_span, left_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return left_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return left_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other) -> bool:
        """
        Returns whether ``self`` is to the left of ``other`` allowing overlap. That is, ``self`` ends before ``other`` ends (or
        at the same time).

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_spanset_span, overleft_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return overleft_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overleft_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other) -> bool:
        """
        Returns whether ``self`` is to the right of ``other`` allowing overlap. That is, ``self`` starts after ``other`` starts
        (or at the same time).

        Args:
            other: temporal object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_spanset_span, overright_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return overright_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overright_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other) -> bool:
        """
        Returns whether ``self`` is strictly to the right of ``other``.That is, ``self`` starts after ``other`` ends.

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            right_spanset_span, right_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return right_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return right_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other: object to compare with

        Returns:
            A :class:`float` instance

        MEOS Functions:
        """
        from .span import Span
        if isinstance(other, Span):
            return distance_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return distance_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Set Operations --------------------------------
    @abstractmethod
    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A collection instance. The actual class depends on ``other``.

        MEOS Functions:
        intersection_spanset_spanset, intersection_spanset_span
        """
        from .span import Span
        if isinstance(other, Span):
            return intersection_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return intersection_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __mul__(self, other):
        """
        Returns the temporal intersection of ``self`` and ``other``.

        Args:
            other: temporal object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
        intersection_periodset_timestamp, intersection_spanset_spanset, intersection_spanset_span
        """
        return self.intersection(other)

    @abstractmethod
    def minus(self, other):
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
        minus_spanset_span, minus_spanset_spanset
        """
        from .span import Span
        if isinstance(other, Span):
            return minus_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return minus_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __sub__(self, other):
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
        minus_spanset_span, minus_spanset_spanset, minus_periodset_timestamp
        """
        return self.minus(other)

    @abstractmethod
    def union(self, other):
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_periodset_timestamp, union_spanset_spanset,
            union_spanset_span
        """
        from .span import Span
        if isinstance(other, Span):
            return union_spanset_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return union_spanset_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __add__(self, other):
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
        union_periodset_timestamp, union_spanset_spanset, union_spanset_span
        """
        return self.union(other)

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Return whether ``self`` and ``other`` are equal.

        Args:
            other: temporal object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            spanset_eq
        """
        if isinstance(other, self.__class__):
            return spanset_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        """
        Return whether ``self`` and ``other`` are not equal.

        Args:
            other: temporal object to compare with

        Returns:
            True if not equal, False otherwise

        MEOS Functions:
            spanset_ne
        """
        if isinstance(other, self.__class__):
            return spanset_ne(self._inner, other._inner)
        return True

    def __lt__(self, other):
        """
        Return whether ``self`` is less than ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than, False otherwise

        MEOS Functions:
            spanset_lt
        """
        if isinstance(other, self.__class__):
            return spanset_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        """
        Return whether ``self`` is less than or equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than or equal, False otherwise

        MEOS Functions:
            spanset_le
        """
        if isinstance(other, self.__class__):
            return spanset_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        """
        Return whether ``self`` is greater than ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than, False otherwise

        MEOS Functions:
            spanset_gt
        """
        if isinstance(other, self.__class__):
            return spanset_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        """
        Return whether ``self`` is greater than or equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than or equal, False otherwise

        MEOS Functions:
            spanset_ge
        """
        if isinstance(other, self.__class__):
            return spanset_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')



from __future__ import annotations

import abc
from datetime import datetime, timedelta
from typing import Optional, Union, List, Generic, TypeVar, Callable

from pymeos_cffi import *

from ..boxes import Box
from ..factory import _TemporalFactory
from ..temporal import Temporal
from ..collections import Time

ResultType = TypeVar('ResultType', bound=Union[Temporal, Time, Box])
SourceType = TypeVar('SourceType', bound=Union[Temporal, Time, Box])
StateType = TypeVar('StateType')
SourceMeosType = TypeVar('SourceMeosType')
ResultMeosType = TypeVar('ResultMeosType')
SelfAgg = TypeVar('SelfAgg', bound='Aggregation')

IntervalType = TypeVar('IntervalType')
OriginType = TypeVar('OriginType')


class BaseAggregator(Generic[SourceType, ResultType], abc.ABC):
    """
    Abstract class for all aggregations.

    Can be easily extended overriding the :func:`~pymeos.aggregators.aggregator.BaseAggregator._add_function` and
    :func:`~pymeos.aggregators.aggregator.BaseAggregator._final_function` when the aggregation uses the inner MEOS
    objects, or :func:`~pymeos.aggregators.aggregator.BaseAggregator._add` and
    :func:`~pymeos.aggregators.aggregator.BaseAggregator._finish` for arbitrary aggregations.
    """

    @staticmethod
    def _add_function(state: Optional[StateType], meos_object: SourceMeosType) -> StateType:
        """
        Add `meos_object` to the aggregation. Usually a MEOS function.
        Args:
            state: current state of the aggregation.
            meos_object: new MEOS object to aggregate.

        Returns:
            New state of the aggregation after adding `meos_object`.

        """
        raise NotImplemented

    @staticmethod
    def _final_function(state: StateType) -> ResultMeosType:
        """
        Return the current value of the aggregation. Usually a MEOS function.
        Args:
            state: current state of the aggregation.

        Returns:
            Value of the aggregation.

        """
        return temporal_tagg_finalfn(state)

    @classmethod
    def _add(cls, state: Optional[StateType], temporal: SourceType) -> StateType:
        """
        Add the `temporal` object to the aggregation.
        Args:
            state: current state of the aggregation.
            temporal: new PyMEOS object to aggregate.

        Returns:
            New state of the aggregation after adding `temporal`.

        """
        return cls._add_function(state, temporal._inner)

    @classmethod
    def _finish(cls, state) -> ResultType:
        """
        Return the current value of the aggregation.
        Args:
            state: current state of the aggregation.

        Returns:
            Value of the aggregation.

        """
        result = cls._final_function(state)
        return _TemporalFactory.create_temporal(result)

    @classmethod
    def aggregate(cls, temporals: List[SourceType]) -> ResultType:
        """
        Aggregate a list of PyMEOS object at once.

        Useful when only the final result is desired and all the objects are available.

        For aggregating in a streaming fashion, see
        :func:'pymeos.aggregators.aggregator.BaseAggregator.start_aggregation'.

        Args:
            temporals: list of PyMEOS objects to aggregate.

        Returns:
            Result of applying the aggregation to the passed objects.

        """
        state = None
        for t in temporals:
            state = cls._add(state, t)
        return cls._finish(state)

    @classmethod
    def start_aggregation(cls) -> Aggregation:
        """
        Return an :class:`Aggregation` object that holds the state of a new aggregation.

        Returns:
            An :class:`Aggregation` instance
        """
        return Aggregation(cls._add, cls._finish)

    @classmethod
    def _error(cls, element):
        raise TypeError(f'Cannot perform aggregation ({cls.__name__}) with the following element: '
                        f'{element} (Class: {element.__class__})')


class Aggregation(Generic[SourceType, ResultType]):
    """
    Class representing an aggregation in process.

    This class is returned by the :func:`~pymeos.aggregators.aggregator.BaseAggregator.start_aggregation` method
    of :class:`BaseAggregator` subclasses, and shouldn't be created directly by the final user.
    """

    def __init__(self, add_function, finish_function) -> None:
        super().__init__()
        self._add_function = add_function
        self._finish_function = finish_function
        self._state = None

    def add(self: SelfAgg, new_temporal: SourceType) -> SelfAgg:
        """
        Add a new element to the aggregation.

        Returns itself to allowing chaining

        Examples:
            >>> aggregation.add(temporal1).add(temporal2)
            Is equivalent to
            >>> aggregation.add(temporal1)
            >>> aggregation.add(temporal2)

        Args:
            new_temporal: PyMEOS object to add to the aggregation

        Returns:
            `self`
        """
        self._state = self._add_function(self._state, new_temporal)
        return self

    def aggregation(self) -> ResultType:
        """
        Return the current aggregation value.

        Note that this doesn't finish the aggregation, so more elements can be still added.

        Examples:
            >>> aggregation.add(temporal1).add(temporal2)
            >>> intermediate = aggregation.aggregation()
            >>> aggregation.add(temporal3).add(temporal4)
            >>> final_result = aggregation.aggregation()

        Returns:
            Result of the aggregation
        """
        return self._finish_function(self._state)


class BaseGranularAggregator(Generic[SourceType, ResultType, IntervalType, OriginType]):
    """
    Abstract class for granular aggregations.
    """

    @staticmethod
    def _add_function(state: Optional[StateType], meos_object: SourceMeosType, interval: IntervalType,
                      origin: OriginType) -> StateType:
        """
        Add `meos_object` to the aggregation. Usually a MEOS function.
        Args:
            state: current state of the aggregation.
            meos_object: new MEOS object to aggregate.
            interval: width of the aggregation intervals
            origin: starting value of the first interval

        Returns:
            New state of the aggregation after adding `meos_object`.

        """
        raise NotImplemented

    @staticmethod
    def _final_function(state: StateType) -> ResultMeosType:
        """
        Return the current value of the aggregation. Usually a MEOS function.
        Args:
            state: current state of the aggregation.

        Returns:
            Value of the aggregation.

        """
        return temporal_tagg_finalfn(state)

    @classmethod
    def _add(cls, state: Optional[StateType], temporal: SourceType, interval: IntervalType,
             origin: OriginType) -> StateType:
        """
        Add the `temporal` object to the aggregation.
        Args:
            state: current state of the aggregation.
            temporal: new PyMEOS object to aggregate.
            interval: width of the aggregation intervals
            origin: starting value of the first interval

        Returns:
            New state of the aggregation after adding `temporal`.

        """
        interval_converted = timedelta_to_interval(interval) if isinstance(interval, timedelta) else \
            pg_interval_in(interval, -1) if isinstance(interval, str) else None
        origin_converted = datetime_to_timestamptz(origin) if isinstance(origin, datetime) else \
            pg_timestamptz_in(origin, -1)
        return cls._add_function(state, temporal._inner, interval_converted, origin_converted)

    @classmethod
    def _finish(cls, state: StateType) -> ResultType:
        """
        Return the current value of the aggregation.
        Args:
            state: current state of the aggregation.

        Returns:
            Value of the aggregation.

        """
        result = cls._final_function(state)
        return _TemporalFactory.create_temporal(result)

    @classmethod
    def aggregate(cls, temporals: List[SourceType], interval: IntervalType, origin: OriginType) -> ResultType:
        """
        Aggregate a list of PyMEOS object at once with certain granularity.

        Useful when only the final result is desired and all the objects are available.

        For aggregating in a streaming fashion, see
        :func:`~pymeos.aggregators.aggregator.BaseGranularAggregator.start_aggregation`.

        Args:
            temporals: list of PyMEOS objects to aggregate.
            interval: width of the aggregation intervals
            origin: starting value of the first interval

        Returns:
            Result of applying the granular aggregation to the passed objects.

        """
        state = None
        for t in temporals:
            state = cls._add(state, t, interval, origin)
        return cls._finish(state)

    @classmethod
    def start_aggregation(cls, interval: IntervalType, origin: OriginType) -> \
            GranularAggregation[SourceType, ResultType, IntervalType, OriginType]:
        return GranularAggregation(cls._add, cls._finish, interval, origin)

    @classmethod
    def _error(cls, element):
        raise TypeError(f'Cannot perform aggregation ({cls.__name__}) with the following element: '
                        f'{element} (Class: {element.__class__})')


class GranularAggregation(Aggregation[SourceType, ResultType],
                          Generic[SourceType, ResultType, IntervalType, OriginType]):
    """
    Class representing a granular aggregation in process.

    This class is returned by the :func:`~pymeos.aggregators.aggregator.BaseGranularAggregator.start_aggregation` method
    of :class:`BaseGranularAggregator` subclasses, and shouldn't be created directly by the final user.
    """

    def __init__(self,
                 add_function: Callable[[Optional[StateType], SourceType, IntervalType, OriginType], StateType],
                 finish_function: Callable[[StateType], ResultType],
                 interval: IntervalType,
                 origin: OriginType) -> None:
        super().__init__(add_function, finish_function)
        self._interval = interval
        self._origin = origin

    def add(self: SelfAgg, new_temporal: SourceType) -> SelfAgg:
        self._state = self._add_function(self._state, new_temporal, self._interval, self._origin)
        return self



from __future__ import annotations

from typing import Optional, Union, List

from pymeos_cffi import *

from ..main import TNumber
from ..collections import *


class TBox:
    """
    Class for representing numeric temporal boxes. Both numeric and temporal
    bounds may be inclusive or not.

    ``TBox`` objects can be created with a single argument of type string as in
    MobilityDB.

        >>> TBox('TBOXINT XT([0, 10),[2020-06-01, 2020-06-05])')
        >>> TBox('TBOXFLOAT XT([0, 10),[2020-06-01, 2020-06-05])')
        >>> TBox('TBOX T([2020-06-01, 2020-06-05])')

    Another possibility is to provide the ``xmin``/``xmax`` (of type str or
    int/float) and ``tmin``/``tmax`` (of type str or datetime) named parameters, 
    and optionally indicate whether the bounds are inclusive or exclusive (by 
    default, lower bounds are inclusive and upper bounds are exclusive):

        >>> TBox(xmin=0, xmax=10, tmin='2020-06-01', tmax='2020-06-0')
        >>> TBox(xmin=0, xmax=10, tmin='2020-06-01', tmax='2020-06-0', xmax_inc=True, tmax_inc=True)
        >>> TBox(xmin='0', xmax='10', tmin=parse('2020-06-01'), tmax=parse('2020-06-0'))

    Note that you can create a TBox with only the numerical or the temporal
    dimension. In these cases, it will be equivalent to a 
    :class:`~pymeos.time.period.Period` (if it only has temporal dimension) or
    to a :class:`FloatSpan` (if it only has the numeric dimension).
    """
    __slots__ = ['_inner']

    _mobilitydb_name = 'tbox'

    def _inner_period(self):
        from pymeos_cffi.functions import _ffi
        return _ffi.addressof(self._inner.period)

    def _inner_span(self):
        from pymeos_cffi.functions import _ffi
        return _ffi.addressof(self._inner.span)

    def _is_float(self) -> bool:
        return self._inner.span.basetype == 5

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 xmin: Optional[Union[str, int, float]] = None,
                 xmax: Optional[Union[str, int, float]] = None,
                 tmin: Optional[Union[str, datetime]] = None,
                 tmax: Optional[Union[str, datetime]] = None,
                 xmin_inc: bool = True,
                 xmax_inc: bool = False,
                 tmin_inc: bool = True,
                 tmax_inc: bool = False,
                 _inner=None):
        assert (_inner is not None) or (string is not None) != (
                (xmin is not None and xmax is not None) or
                (tmin is not None and tmax is not None)), \
            "Either string must be not None or at least a bound pair (xmin/max or tmin/max) must be not None"
        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = tbox_in(string)
        else:
            span = None
            period = None
            if xmin is not None and xmax is not None:
                if isinstance(xmin, int) and isinstance(xmax, int):
                    span = intspan_make(xmin, xmax, xmin_inc, xmax_inc)
                else:
                    span = floatspan_make(float(xmin), float(xmax), xmin_inc,
                                          xmax_inc)
            if tmin is not None and tmax is not None:
                period = Period(lower=tmin, upper=tmax, lower_inc=tmin_inc,
                                upper_inc=tmax_inc)._inner
            self._inner = tbox_make(span, period)

    def __copy__(self) -> TBox:
        """
        Returns a copy of ``self``.

        Returns:
            A :class:`TBox` instance.

        MEOS Functions:
            tbox_copy
        """
        inner_copy = tbox_copy(self._inner)
        return TBox(_inner=inner_copy)

    @staticmethod
    def from_wkb(wkb: bytes) -> TBox:
        """
        Returns a `TBox` from its WKB representation.

        Args:
            wkb: WKB representation

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            tbox_from_wkb
        """
        result = tbox_from_wkb(wkb)
        return TBox(_inner=result)

    @staticmethod
    def from_hexwkb(hexwkb: str) -> TBox:
        """
        Returns a `TBox` from its WKB representation in hex-encoded ASCII.

        Args:
            hexwkb: WKB representation in hex-encoded ASCII

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            tbox_from_hexwkb
        """
        result = tbox_from_hexwkb(hexwkb)
        return TBox(_inner=result)

    @staticmethod
    def from_value(value: Union[int, float, IntSpan, FloatSpan]) -> TBox:
        """
        Returns a `TBox` from a numeric value or span. The created `TBox` will
        only have a numerical dimension.

        Args:
            value: value to be canverted into a TBox

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            int_to_tbox, float_to_tbox, span_to_tbox, span_to_tbox
        """
        if isinstance(value, int):
            result = int_to_tbox(value)
        elif isinstance(value, float):
            result = float_to_tbox(value)
        elif isinstance(value, IntSpan):
            result = numspan_to_tbox(value._inner)
        elif isinstance(value, FloatSpan):
            result = numspan_to_tbox(value._inner)
        else:
            raise TypeError(f'Operation not supported with type {value.__class__}')
        return TBox(_inner=result)

    @staticmethod
    def from_time(time: Time) -> TBox:
        """
        Returns a `TBox` from a :class:`~pymeos.time.time.Time` object. The
        created `TBox` will only have a temporal dimension.

        Args:
            time: value to be canverted into a TBox

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            timestamp_to_tbox, timestampset_to_tbox, period_to_tbox,
            periodset_to_tbox
        """
        if isinstance(time, datetime):
            result = timestamp_to_tbox(datetime_to_timestamptz(time))
        elif isinstance(time, TimestampSet):
            result = timestampset_to_tbox(time._inner)
        elif isinstance(time, Period):
            result = period_to_tbox(time._inner)
        elif isinstance(time, PeriodSet):
            result = periodset_to_tbox(time._inner)
        else:
            raise TypeError(f'Operation not supported with type {time.__class__}')
        return TBox(_inner=result)

    @staticmethod
    def from_value_time(value: Union[int, float, IntSpan, FloatSpan],
                        time: Union[datetime, Period]) -> TBox:
        """
        Returns a `TBox` from a numerical and a temporal object.

        Args:
            value: numerical span of the TBox
            time: temporal span of the TBox

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            int_timestamp_to_tbox, int_period_to_tbox, 
            float_timestamp_to_tbox, float_period_to_tbox,
            span_timestamp_to_tbox, span_period_to_tbox
        """
        if isinstance(value, int) and isinstance(time, datetime):
            result = int_timestamp_to_tbox(value,
                                           datetime_to_timestamptz(time))
        elif isinstance(value, int) and isinstance(time, Period):
            result = int_period_to_tbox(value, time._inner)
        elif isinstance(value, float) and isinstance(time, datetime):
            result = float_timestamp_to_tbox(value,
                                             datetime_to_timestamptz(time))
        elif isinstance(value, float) and isinstance(time, Period):
            result = float_period_to_tbox(value, time._inner)
        elif isinstance(value, IntSpan) and isinstance(time, datetime):
            result = span_timestamp_to_tbox(value._inner,
                                            datetime_to_timestamptz(time))
        elif isinstance(value, IntSpan) and isinstance(time, Period):
            result = span_period_to_tbox(value._inner, time._inner)
        elif isinstance(value, FloatSpan) and isinstance(time, datetime):
            result = span_timestamp_to_tbox(value._inner,
                                            datetime_to_timestamptz(time))
        elif isinstance(value, FloatSpan) and isinstance(time, Period):
            result = span_period_to_tbox(value._inner, time._inner)
        else:
            raise TypeError(f'Operation not supported with types {value.__class__} and {time.__class__}')
        return TBox(_inner=result)

    @staticmethod
    def from_tnumber(temporal: TNumber) -> TBox:
        """
        Returns a `TBox` from a :class:`~pymeos.main.tnumber.TNumber` object.

        Args:
            temporal: temporal number to be canverted into a TBox

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            tnumber_to_tbox
        """
        return TBox(_inner=tnumber_to_tbox(temporal._inner))

    # ------------------------- Output ----------------------------------------
    def __str__(self, max_decimals: int = 15):
        """
        Returns a string representation of ``self``.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            tbox_out
        """
        return tbox_out(self._inner, max_decimals)

    def __repr__(self):
        """
        Returns a string representation of ``self``.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            tbox_out
        """
        return (f'{self.__class__.__name__}'
                f'({self})')

    def as_wkb(self) -> bytes:
        """
        Returns the WKB representation of ``self``.

        Returns:
            A :class:`str` object with the WKB representation of ``self``.

        MEOS Functions:
            tbox_as_wkb
        """
        return tbox_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the WKB representation of ``self`` in hex-encoded ASCII.

        Returns:
            A :class:`str` object with the WKB representation of ``self`` in
            hex-encoded ASCII.

        MEOS Functions:
            tbox_as_hexwkb
        """
        return tbox_as_hexwkb(self._inner, -1)[0]

    # ------------------------- Conversions ----------------------------------
    def to_floatspan(self) -> FloatSpan:
        """
        Returns the numeric span of ``self``.

        Returns:
            A new :class:`FloatSpan` instance

        MEOS Functions:
            tbox_to_floatspan
        """
        from ..collections import FloatSpan
        return FloatSpan(_inner=tbox_to_floatspan(self._inner))

    def to_period(self) -> Period:
        """
        Returns the temporal span of ``self``.

        Returns:
            A new :class:`~pymeos.time.period.Period` instance

        MEOS Functions:
            tbox_to_period
        """
        return Period(_inner=tbox_to_period(self._inner))

    # ------------------------- Accessors -------------------------------------
    def has_x(self) -> bool:
        """
        Returns whether ``self`` has a numeric dimension.

        Returns:
            True if ``self`` has a numeric dimension, False otherwise

        MEOS Functions:
            tbox_hasx
        """
        return tbox_hasx(self._inner)

    def has_t(self) -> bool:
        """
        Returns whether ``self`` has a temporal dimension.

        Returns:
            True if ``self`` has a temporal dimension, False otherwise

        MEOS Functions:
            tbox_hast
        """
        return tbox_hast(self._inner)

    def xmin(self) -> float:
        """
        Returns the numeric lower bound of ``self``.

        Returns:
            The numeric lower bound of the `TBox` as a :class:`float`

        MEOS Functions:
            tbox_xmin
        """
        return tbox_xmin(self._inner)

    def xmin_inc(self) -> bool:
        """
        Returns whether the xmin value of the tbox is inclusive or not

        Returns:
            True if the xmin value of the tbox is inclusive and False otherwise

        MEOS Functions:
            tbox_xmin_inc
        """
        return tbox_xmin_inc(self._inner)

    def xmax(self) -> float:
        """
        Returns the numeric upper bound of ``self``.

        Returns:
            The numeric upper bound of the `TBox` as a :class:`float`

        MEOS Functions:
            tbox_xmax
        """
        return tbox_xmax(self._inner)

    def xmax_inc(self) -> bool:
        """
        Returns whether the xmax value of the tbox is inclusive or not

        Returns:
            True if the xmax value of the tbox is inclusive and False otherwise

        MEOS Functions:
            tbox_xmax_inc
        """
        return tbox_xmax_inc(self._inner)

    def tmin(self):
        """
        Returns the temporal lower bound of ``self``.

        Returns:
            The temporal lower bound of the `TBox` as a 
            :class:`~datetime.datetime`

        MEOS Functions:
            tbox_tmin
        """
        result = tbox_tmin(self._inner)
        if not result:
            return None
        return timestamptz_to_datetime(result)

    def tmin_inc(self) -> bool:
        """
        Returns whether the tmin value of the tbox is inclusive or not

        Returns:
            True if the tmin value of the tbox is inclusive and False otherwise

        MEOS Functions:
            tbox_tmin_inc
        """
        return tbox_tmin_inc(self._inner)

    def tmax(self):
        """
        Returns the temporal upper bound of ``self``.

        Returns:
            The temporal upper bound of the `TBox` as a 
            :class:`~datetime.datetime`

        MEOS Functions:
            tbox_tmax
        """
        result = tbox_tmax(self._inner)
        if not result:
            return None
        return timestamptz_to_datetime(result)

    def tmax_inc(self) -> bool:
        """
        Returns whether the tmax value of the tbox is inclusive or not

        Returns:
            True if the tmax value of the tbox is inclusive and False otherwise

        MEOS Functions:
            tbox_tmax_inc
        """
        return tbox_tmax_inc(self._inner)

    # ------------------------- Transformation --------------------------------
    def expand(self, other: Union[int, float, timedelta]) -> TBox:
        """
        Returns the result of expanding ``self`` with the ``other``. Depending
        on the type of ``other``, the expansion will be of the numeric
        dimension (:class:`float`) or temporal (:class:`~datetime.timedelta`).

        Args:
            other: object used to expand ``self``

        Returns:
            A new :class:`TBox` instance.

        MEOS Functions:
            tbox_expand_value, tbox_expand_time
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tbox_expand_value(self._inner, float(other))
        elif isinstance(other, timedelta):
            result = tbox_expand_time(self._inner, timedelta_to_interval(other))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return TBox(_inner=result)

    def shift_value(self, delta: Union[int, float]) -> TBox:
        """
        Returns a new `TBox` with the value dimension shifted by `delta`.

        Args:
            delta: value to shift

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            span_shift_scale

        See Also:
            :meth:`Span.shift`
        """
        return self.shift_scale_value(shift=delta)

    def shift_time(self, delta: timedelta) -> TBox:
        """
        Returns a new `TBox` with the time dimension shifted by `delta`.

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            period_shift_scale

        See Also:
            :meth:`Period.shift`
        """
        return self.shift_scale_time(shift=delta)

    def scale_value(self, width: Union[int, float]) -> TBox:
        """
        Returns a new `TBox` with the value dimension having width `width`.

        Args:
            width: value of the new width

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            span_shift_scale

        See Also:
            :meth:`Span.scale`
        """
        return self.shift_scale_value(width=width)

    def scale_time(self, duration: timedelta) -> TBox:
        """
        Returns a new `TBox` with the time dimension having duration `duration`.

        Args:
            duration: :class:`datetime.timedelta` instance with new duration

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            period_shift_scale

        See Also:
            :meth:`Period.scale`
        """
        return self.shift_scale_time(duration=duration)

    def shift_scale_value(self, shift: Optional[Union[int, float]] = None,
                          width: Optional[Union[int, float]] = None) -> TBox:
        """
        Returns a new TBox with the value span shifted by `shift` and
        width `width`.

        Examples:
            >>> tbox = TBox('TBoxInt XT([0, 5),[2020-06-01, 2020-06-02])')
            >>> tbox.shift_scale_value(shift=2, width=4)
            >>> 'TBOXINT XT([2, 7),[2020-06-01 00:00:00+02, 2020-06-02 00:00:00+02])'

        Args:
            shift: :value to shift the start of the value span
            width: value representing the width of the value span

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            span_shift_scale

        See Also:
            :meth:`Span.shift_scale`
        """
        assert shift is not None or width is not None, \
            'shift and width deltas must not be both None'
        hasshift = shift is not None
        haswidth = width is not None
        if (shift is None or isinstance(shift, int)) and \
                (width is None or isinstance(width, int)):
            result = tbox_shift_scale_int(self._inner,
                                          shift if shift else 0, width if width else 0,
                                          hasshift, haswidth)
        elif (shift is None or isinstance(shift, float)) and \
                (width is None or isinstance(width, float)):
            result = tbox_shift_scale_float(self._inner,
                                            shift if shift else 0.0, width if width else 0.0,
                                            hasshift, haswidth)
        else:
            raise TypeError(f'Operation not supported with type {self.__class__}')
        return TBox(_inner=result)

    def shift_scale_time(self, shift: Optional[timedelta] = None,
                         duration: Optional[timedelta] = None) -> TBox:
        """
        Returns a new TBox with the temporal span shifted by `shift` and
        duration `duration`.

        Examples:
            >>> tbox = TBox('TBoxInt XT([0, 10),[2020-06-01, 2020-06-05])')
            >>> tbox.shift_scale_time(shift=timedelta(days=2), duration=timedelta(days=4))
            >>> 'TBOXINT XT([0, 10),[2020-06-03 00:00:00+02, 2020-06-07 00:00:00+02])'

        Args:
            shift: :class:`datetime.timedelta` instance to shift the start of
                the temporal span
            duration: :class:`datetime.timedelta` instance representing the
                duration of the temporal span

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            period_shift_scale

        See Also:
            :meth:`Period.shift_scale`
        """
        assert shift is not None or duration is not None, \
            'shift and duration deltas must not be both None'
        result = tbox_shift_scale_time(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None
        )
        return TBox(_inner=result)

    def round(self, maxdd: int = 0) -> TBox:
        """
        Returns `self` rounded to the given number of decimal digits.

        Args:
            maxdd: Maximum number of decimal digits.

        Returns:
            A new :class:`TBox` instance

        MEOS Functions:
            tbox_round
        """
        new_inner = tbox_copy(self._inner)
        tbox_round(new_inner, maxdd)
        return TBox(_inner=new_inner)

    # ------------------------- Set Operations --------------------------------
    def union(self, other: TBox, strict: Optional[bool] = True) -> TBox:
        """
        Returns the union of `self` with `other`.

        Args:
            other: temporal object to merge with
            strict: Whether to fail if the boxes do not intersect.

        Returns:
            A :class:`TBox` instance.

        MEOS Functions:
            union_tbox_tbox
        """
        return TBox(_inner=union_tbox_tbox(self._inner, other._inner, strict))

    def __add__(self, other):
        """
        Returns the union of `self` with `other`. Fails if the union is not
        contiguous.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`TBox` instance.

        MEOS Functions:
            union_tbox_tbox
        """
        return self.union(other, False)

    # TODO: Check returning None for empty intersection is the desired behaviour
    def intersection(self, other: TBox) -> Optional[TBox]:
        """
        Returns the intersection of `self` with `other`.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`TBox` instance if the instersection is not empty, `None`
            otherwise.

        MEOS Functions:
            intersection_tbox_tbox
        """
        result = intersection_tbox_tbox(self._inner, other._inner)
        return TBox(_inner=result) if result else None

    def __mul__(self, other):
        """
        Returns the intersection of `self` with `other`.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`TBox` instance if the instersection is not empty, `None`
            otherwise.

        MEOS Functions:
            intersection_tbox_tbox
        """
        return self.intersection(other)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self,
                    other: Union[int, float, IntSpan, FloatSpan, TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share
        only the temporal or numerical bound and only one of them contains it.

        Examples:
            >>> TBox('TBoxInt XT([0, 1], [2012-01-01, 2012-01-02))').is_adjacent(TBox('TBoxInt XT([0, 1], [2012-01-02, 2012-01-03])'))
            >>> True
            >>> TBox('TBoxInt XT([0, 1], [2012-01-01, 2012-01-02])').is_adjacent(TBox('TBoxInt XT([0, 1], [2012-01-02, 2012-01-03])'))
            >>> False  # Both contain bound
            >>> TBox('TBoxInt XT([0, 1), [2012-01-01, 2012-01-02))').is_adjacent(TBox('TBoxInt XT([1, 2], [2012-01-02, 2012-01-03])')
            >>> False  # Adjacent in both bounds

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, int):
            return adjacent_span_span(self._inner_span(),
                                      float_to_floatspan(float(other)))
        elif isinstance(other, float):
            return adjacent_span_span(self._inner_span(),
                                      float_to_floatspan(other))
        elif isinstance(other, IntSpan):
            from pymeos_cffi.functions import _ffi
            return adjacent_span_span(_ffi.addressof(self._inner, 'span'),
                                      other._inner)
        elif isinstance(other, FloatSpan):
            return adjacent_span_span(self._inner.span, other._inner)
        elif isinstance(other, TBox):
            return adjacent_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return adjacent_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_contained_in(self, container: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is contained in ``container``.

        Examples:
            >>> TBox('TBoxInt XT([1, 2], [2012-01-02, 2012-01-03])').is_contained_in(TBox('TBoxInt XT([1, 4], [2012-01-01, 2012-01-04])'))
            >>> True
            >>> TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02))').is_contained_in(TBox('TBoxFloat XT([1, 4], [2012-01-01, 2012-01-02])'))
            >>> True
            >>> TBox('TBoxFloat XT([1, 2], [2012-01-01, 2012-01-02])').is_contained_in(TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02))'))
            >>> False

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(container, TBox):
            return contained_tbox_tbox(self._inner, container._inner)
        elif isinstance(container, TNumber):
            return contained_tbox_tbox(self._inner,
                                       tnumber_to_tbox(container._inner))
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')

    def contains(self, content: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> TBox('TBoxInt XT([1, 4], [2012-01-01, 2012-01-04]').contains(TBox('TBoxInt XT([2, 3], [2012-01-02, 2012-01-03]'))
            >>> True
            >>> TBox('TBoxFloat XT([1, 2], [2012-01-01, 2012-01-02]').contains(TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02)'))
            >>> True
            >>> TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02)').contains(TBox('TBoxFloat XT([1, 2], [2012-01-01, 2012-01-02]'))
            >>> False

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(content, TBox):
            return contains_tbox_tbox(self._inner, content._inner)
        elif isinstance(content, TNumber):
            return contains_tbox_tbox(self._inner,
                                      tnumber_to_tbox(content._inner))
        else:
            raise TypeError(f'Operation not supported with type {content.__class__}')

    def __contains__(self, item):
        """
        Returns whether ``self`` temporally contains ``item``.

        Examples:
            >>> TBox('TBoxInt XT([2, 3], [2012-01-02, 2012-01-03]') in TBox('TBoxInt XT([1, 4], [2012-01-01, 2012-01-04]')
            >>> True
            >>> TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02)') in TBox('TBoxFloat XT([1, 2], [2012-01-01, 2012-01-02]')
            >>> True
            >>> TBox('TBoxFloat XT([1, 2], [2012-01-01, 2012-01-02]') in TBox('TBoxFloat XT((1, 2), (2012-01-01, 2012-01-02)')
            >>> False

        Args:
            item: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_tbox_tbox, tnumber_to_tbox
        """
        return self.contains(item)

    def overlaps(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` overlaps ``other``. That is, both share at
        least an instant or a value.

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return overlaps_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overlaps_tbox_tbox(self._inner,
                                      tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_same(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is the same as ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if same, False otherwise

        MEOS Functions:
            same_tbox_tbox, tnumber_to_tbox

        """
        if isinstance(other, TBox):
            return same_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return same_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if left, False otherwise

        MEOS Functions:
            left_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return left_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return left_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is to the left of ``other`` allowing overlap.
        That is, ``self`` does not extend to the right of ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if overleft, False otherwise

        MEOS Functions:
            overleft_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return overleft_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overleft_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is strictly to the right of ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if right, False otherwise

        MEOS Functions:
            right_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return right_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return right_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is to the right of ``other`` allowing overlap.
        That is, ``self`` does not extend to the left of ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if overright, False otherwise

        MEOS Functions:
            overright_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return overright_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overright_tbox_tbox(self._inner,
                                       tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_before(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            before_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return before_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return before_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_before(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` does not extend after ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if overbefore, False otherwise

        MEOS Functions:
            overbefore_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return overbefore_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overbefore_tbox_tbox(self._inner,
                                        tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_after(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            after_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return after_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return after_tbox_tbox(self._inner, tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_after(self, other: Union[TBox, TNumber]) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` does not extend before``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if overafter, False otherwise

        MEOS Functions:
            overafter_tbox_tbox, tnumber_to_tbox
        """
        if isinstance(other, TBox):
            return overafter_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overafter_tbox_tbox(self._inner,
                                       tnumber_to_tbox(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Distance Operations ---------------------------
    def nearest_approach_distance(self, other: Union[TBox, TNumber]) -> float:
        """
        Returns the distance between the nearest points of ``self`` and ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            A :class:`float` with the distance between the nearest points of
            ``self`` and ``other``.

        MEOS Functions:
            nad_tbox_tbox
        """
        if isinstance(other, TBox):
            return nad_tbox_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return nad_tnumber_tbox(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Splitting --------------------------------------
    def tile(self, size: float, duration: Union[timedelta, str],
             origin: float = 0.0, start: Union[datetime, str, None] = None) -> \
            List[TBox]:
        """
        Returns 2d matrix of TBoxes resulting of tiling ``self``.

        Args:
            size: size of the numeric dimension of the tiles
            duration: size of the temporal dimenstion of the tiles
            origin: origin of the numeric dimension of the tiles
            start: origin of the temporal dimension of the tiles. If None, the
                start time used by default is Monday, January 3, 2000.

        Returns:
            An array of :class:`TBox` instances.

        MEOS Functions:
            tintbox_tile_list, tfloabox_tile_list
        """
        dt = timedelta_to_interval(duration) if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1)
        st = datetime_to_timestamptz(start) if isinstance(start, datetime) \
            else pg_timestamptz_in(start, -1) if isinstance(start, str) \
            else pg_timestamptz_in('2000-01-03', -1)
        if self._is_float():
            tiles, count = tfloatbox_tile_list(self._inner, size, dt, origin, st)
        else:
            tiles, count = tintbox_tile_list(self._inner, int(size), dt, int(origin), st)
        return [TBox(_inner=tiles + c) for c in range(count)]

    # def tile_flat(self, size: float, duration: Union[timedelta, str],
    #               origin: float = 0.0,
    #               start: Union[datetime, str, None] = None) -> List[TBox]:
    #     """
    #     Returns an array of TBoxes resulting of tiling ``self``.
    #
    #     Args:
    #         size: size of the numeric dimension of the tiles
    #         duration: size of the temporal dimenstion of the tiles
    #         origin: origin of the numeric dimension of the tiles
    #         start: origin of the temporal dimension of the tiles. If None, the
    #             start time used by default is Monday, January 3, 2000.
    #
    #     Returns:
    #         An array of :class:`TBox` instances.
    #
    #     MEOS Functions:
    #         tbox_tile_list
    #     """
    #     tiles = self.tile(size, duration, origin, start)
    #     return [box for row in tiles for box in row]

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Returns whether ``self`` is equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            tbox_eq
        """
        if isinstance(other, self.__class__):
            return tbox_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        """
        Returns whether ``self`` is not equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if not equal, False otherwise

        MEOS Functions:
            tbox_ne
        """
        if isinstance(other, self.__class__):
            return tbox_ne(self._inner, other._inner)
        return True

    def __lt__(self, other):
        """
        Returns whether ``self`` is less than ``other``. The time
        dimension is compared first, then the space dimension.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than, False otherwise

        MEOS Functions:
            tbox_lt
        """
        if isinstance(other, self.__class__):
            return tbox_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        """
        Returns whether ``self`` is less than or equal to ``other``. The time
        dimension is compared first, then the space dimension.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than or equal to, False otherwise

        MEOS Functions:
            tbox_le
        """
        if isinstance(other, self.__class__):
            return tbox_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        """
        Returns whether ``self`` is greater than ``other``. The time dimension
        is compared first, then the space dimension.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than, False otherwise

        MEOS Functions:
            tbox_gt
        """
        if isinstance(other, self.__class__):
            return tbox_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        """
        Returns whether ``self`` is greater than or equal to ``other``. The
        time dimension is compared first, then the space dimension.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than or equal to, False otherwise

        MEOS Functions:
            tbox_ge
        """
        if isinstance(other, self.__class__):
            return tbox_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        """
        Plots ``self``.

        See Also:
            :func:`~pymeos.plotters.box_plotter.BoxPlotter.plot_tbox`
        """
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_tbox(self, *args, **kwargs)

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TBox` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        return TBox(string=value)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar('T')
Self = TypeVar('Self', bound='Set[Any]')


class Collection(Generic[T], ABC):
    # ------------------------- Topological Operations ------------------------
    # @abstractmethod
    # def is_adjacent(self, other) -> bool:
    # raise NotImplementedError()

    @abstractmethod
    def is_contained_in(self, container) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def contains(self, content) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError()

    @abstractmethod
    def overlaps(self, other) -> bool:
        raise NotImplementedError()

    # @abstractmethod
    # def is_same(self, other) -> bool:
    # raise NotImplementedError()

    # ------------------------- Position Operations ---------------------------
    @abstractmethod
    def is_left(self, other) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def is_over_or_left(self, other) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def is_over_or_right(self, other) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def is_right(self, other) -> bool:
        raise NotImplementedError()

    # ------------------------- Database Operations ---------------------------

    # ------------------------- Database Operations ---------------------------
    @classmethod
    def read_from_cursor(cls, value, _=None):
        """
        Reads a :class:`STBox` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        return cls(string=value)



from typing import Union

from .stbox import STBox
from .tbox import TBox

Box = Union[TBox, STBox]
"""
Union type that includes all Box types in PyMEOS:  

- :class:`~pymeos.boxes.tbox.TBox` for numeric temporal boxes.  
- :class:`~pymeos.boxes.stbox.STBox` for spatio-temporal boxes.
"""



from datetime import datetime

from matplotlib import pyplot as plt

from ..collections import TimestampSet, Period, PeriodSet


class TimePlotter:
    """
    Plotter for :class:`Time` objects.
    """

    @staticmethod
    def plot_timestamp(timestamp: datetime, *args, axes=None, **kwargs):
        """
        Plot a :class:`datetime` on the given axes as a vertical line.

        Params:
            timestamp: The :class:`datetime` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        return base.axvline(timestamp, *args, **kwargs)

    @staticmethod
    def plot_timestampset(timestampset: TimestampSet, *args, axes=None, **kwargs):
        """
        Plot a :class:`TimestampSet` on the given axes as a vertical line for each timestamp.

        Params:
            timestampset: The :class:`TimestampSet` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        stamps = timestampset.timestamps()
        plot = base.axvline(stamps[0], *args, **kwargs)
        kwargs.pop('label', None)
        plots = [plot]
        for stamp in stamps[1:]:
            plots.append(base.axvline(stamp, *args, color=plot.get_color(), **kwargs))
        return plots

    @staticmethod
    def plot_period(period: Period, *args, axes=None, **kwargs):
        """
        Plot a :class:`Period` on the given axes as two vertical lines filled in between. The lines will be
        dashed if the period is open.

        Params:
            period: The :class:`Period` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        ll = base.axvline(period.lower(), *args, linestyle='-' if period.lower_inc() else '--', **kwargs)
        kwargs.pop('label', None)
        ul = base.axvline(period.upper(), *args, linestyle='-' if period.upper_inc() else '--', **kwargs)
        s = base.axvspan(period.lower(), period.upper(), *args, alpha=0.3, **kwargs)
        return [ll, ul, s]

    @staticmethod
    def plot_periodset(periodset: PeriodSet, *args, axes=None, **kwargs):
        """
        Plot a :class:`PeriodSet` on the given axes as a vertical line for each timestamp.

        Params:
            periodset: The :class:`PeriodSet` to plot.
            *args: Additional arguments to pass to the plot function.
            axes: The axes to plot on. If None, the current axes are used.

        Returns:
            List with the plotted elements.

        See also:
            :func:`~pymeos.plotters.time_plotter.TimePlotter.plot_period`
        """
        periods = periodset.periods()
        line = TimePlotter.plot_period(periods[0], *args, axes=axes, **kwargs)
        kwargs.pop('label', None)
        lines = [line]
        if 'color' not in kwargs:
            kwargs['color'] = line[0].get_color()
        for p in periods[1:]:
            lines.append(TimePlotter.plot_period(p, *args, axes=axes, **kwargs))
        return lines



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..main import TBool


class TemporalAndAggregator(BaseAggregator[TBool, TBool]):
    """
    Temporal "and" of aggregated temporal booleans.

    MEOS Functions:
        tbool_tand_transfn, temporal_tagg_finalfn
    """
    _add_function = tbool_tand_transfn


class TemporalOrAggregator(BaseAggregator[TBool, TBool]):
    """
    Temporal "or" of aggregated temporal booleans.

    MEOS Functions:
        tbool_tor_transfn, temporal_tagg_finalfn
    """
    _add_function = tbool_tor_transfn



from .bool_aggregators import *
from .general_aggregators import *
from .number_aggregators import *
from .text_aggregators import *
from .point_aggregators import *
from .time_aggregators import *

__all__ = [
    # General
    'TemporalInstantCountAggregator', 'TemporalPeriodCountAggregator', 'TemporalExtentAggregator',
    # Bool
    'TemporalAndAggregator', 'TemporalOrAggregator',
    # Number
    'TemporalAverageAggregator',
    'TemporalNumberExtentAggregator',
    'TemporalIntMaxAggregator', 'TemporalIntMinAggregator', 'TemporalIntSumAggregator',
    'TemporalFloatMaxAggregator', 'TemporalFloatMinAggregator', 'TemporalFloatSumAggregator',
    # Text
    'TemporalTextMaxAggregator', 'TemporalTextMinAggregator',
    # Point
    'TemporalPointExtentAggregator',
    # Time
    'TimeInstantaneousUnionAggregator', 'TimeContinuousUnionAggregator',
]
